import logging
import collections.abc as collections
from dataclasses import dataclass
from typing import AnyStr, Mapping, Optional, Sequence
from pydantic import BaseModel, Field

import pymongo
from pymongo import MongoClient, UpdateOne, DeleteOne

from index_registry.mongodb.orgsqs_query_entry import QueryEntry, QueryEntries, now_ms
from index_registry.utils.config import Config
from pscr_common.utils import get_utcnow_with_tzinfo, datetime_to_ms_timestamp


_logger = logging.getLogger(__name__)


@dataclass
class GetEntriesResponse:
    entries: Sequence[Mapping]


class IndexTimestamps(BaseModel):
    min_backend_time: Optional[int] = 0
    max_backend_time: Optional[int] = Field(default_factory=now_ms)
    min_device_time: Optional[int] = 0
    max_device_time: Optional[int] = Field(default_factory=now_ms)


class ItemDoesNotExist(Exception):
    pass


class DbManager:
    def __init__(self, config: Config):
        self.config = config
        host = config.get("mongo", "host")
        port = config.get("mongo", "port")
        user = config.get("mongo", "user")
        password = config.get("mongo", "password")
        db_name = config.get("mongo", "db_name")
        collection_name = config.get("org_sqs", "mongodb_collection")
        direct_connect = config.getboolean("mongo", "direct_connect")
        local_dev_db = config.getboolean("mongo", "local_dev_db")
        if user == "none":
            uri = f"mongodb://{host}:{port}"
        elif direct_connect:
            uri = f"mongodb://{user}:{password}@{host}:{port}/{db_name}?directConnection=true"
        else:
            uri = f"mongodb://{user}:{password}@{host}:{port}/{db_name}"
        client = MongoClient(uri)
        self.db = client[db_name]
        self.db_admin = client.admin
        self.ack_client = self.db[collection_name]
        # If running locally attempt to initialize the DB, this happens with a separate job in prod.
        if local_dev_db:
            self.init_db()

    def init_db(self):
        self.ack_client.create_index([("consumer_name", pymongo.ASCENDING)], unique=True)
        self.ack_client.create_index([("org_key", pymongo.ASCENDING)], unique=False)

    def delete_db(self):
        # Added for testing ease
        self.ack_client.drop()

    def add_entry(self, query: QueryEntry, insert_only_if_exists=False):
        data = query.dict()
        post = {"$set": data}
        query = {"consumer_name": data['consumer_name']}
        ret = self.ack_client.update_one(query, post, not insert_only_if_exists)
        if ret.upserted_id:
            _logger.info(f"Added {data['consumer_name']}")
        elif ret.modified_count:
            _logger.info(f"Updated {data['consumer_name']}")
        elif insert_only_if_exists:
            _logger.info(f"Ignored {data['consumer_name']} row does not exist and insert only if exists is {str(insert_only_if_exists)}")
            raise ItemDoesNotExist(f"Entry {query} does not already exist")
        else:
            _logger.info(f"Ignored {data['consumer_name']} row not added or updated and insert only if exists is {str(insert_only_if_exists)}")

    def add_entry_batch(self, query_entries: QueryEntries):
        request_items = []
        for entry in query_entries.root:
            data = entry.dict()
            query = {"org_key": data['org_key'], "consumer_name": data['consumer_name']}
            post = {"$set": data}
            request_items.append(UpdateOne(query, post, upsert=True))
        self._batch_request(request_items)

    def _batch_request(self, request_items):
        ret = self.ack_client.bulk_write(request_items)
        if not ret.acknowledged:
            raise Exception("Failed to batch update")

    def _generate_query_expression(self, index_timestamps: IndexTimestamps, raw_recent_time: int):
        _logger.debug(f"Passed index_timestamps {index_timestamps}")

        if not raw_recent_time:
            raw_index_range_minutes = self.config.getint("org_sqs", "raw_index_range_minutes")
            now = get_utcnow_with_tzinfo(truncate_to_ms=True)
            raw_recent_time = int(datetime_to_ms_timestamp(now) - (raw_index_range_minutes * 60 * 1000))
        else:
            raw_recent_time = int(raw_recent_time)

        expression = {"$or": [{"$and": [{"type": {"$eq": "raw"}},
                                        {"timestamp": {"$gt": raw_recent_time}},
                                        {"min_backend_timestamp": {"$lt": index_timestamps.max_backend_time}},
                                        {"max_backend_timestamp": {"$gt": index_timestamps.min_backend_time}}]},
                              {"$and": [{"type": {"$ne": "raw"}},
                                        {"min_backend_timestamp": {"$lt": index_timestamps.max_backend_time}},
                                        {"max_backend_timestamp": {"$gt": index_timestamps.min_backend_time}},
                                        {"min_device_timestamp": {"$lt": index_timestamps.max_device_time}},
                                        {"max_device_timestamp": {"$gt": index_timestamps.min_device_time}}]}]}
        return expression

    def get_entries(
        self, org_key: AnyStr, index_timestamps: IndexTimestamps = None,
            raw_recent_time: int = None, count: bool = False) -> GetEntriesResponse | int:
        _logger.debug(
            f"Retrieving query entries associated with {org_key}\n"
            f"{('Index Timestamps: ' + str(index_timestamps)) if index_timestamps else ''}"
        )
        if index_timestamps is not None:
            query = {"$and": [{"org_key": {"$eq": org_key}},
                              self._generate_query_expression(index_timestamps, raw_recent_time)]}
        else:
            query = {"org_key": org_key}
        if count:
            return self.ack_client.count_documents(query)
        else:
            projection = {"_id": 0}
            entries = list(self.ack_client.find(query, projection))
            return GetEntriesResponse(entries=entries)

    def get_entries_count(self, org_key: AnyStr, index_timestamps: IndexTimestamps = None,
                          raw_recent_time: int = None) -> int:
        _logger.debug(
            f"Retrieving count query entries associated with {org_key}\n"
            f"{('Index Timestamps: ' + str(index_timestamps)) if index_timestamps else ''}"
        )
        return self.get_entries(org_key, index_timestamps, raw_recent_time, count=True)

    def iterate_all_entries(self) -> collections.Iterable:
        _logger.debug(f"Iterate all entries")
        projection = {"_id": 0}
        for doc in self.ack_client.find(projection=projection):
            yield QueryEntry.parse_obj(doc)

    def get_entry(self, org_key: AnyStr, consumer_name: AnyStr) -> Optional[QueryEntry]:
        query = {"consumer_name": consumer_name, "org_key": org_key}
        projection = {"_id": 0}
        ret = self.ack_client.find_one(query, projection)
        return QueryEntry.parse_obj(ret) if ret else None

    def delete_entry(self, org_key: AnyStr, consumer_name: AnyStr):
        query = {"org_key": org_key, "consumer_name": consumer_name}
        self.ack_client.delete_one(query)
        _logger.info(f"Deleted entry {consumer_name} org_key {org_key}")

    def delete_entry_batch(self, delete_entries: list):
        request_items = []
        for entry in delete_entries:
            query = {"org_key": entry["org_key"], "consumer_name": entry["consumer_name"]}
            request_items.append(DeleteOne(query))
        self._batch_request(request_items)

    def delete_expired_raw(self, raw_index_timeout_minutes: int):
        now = get_utcnow_with_tzinfo(truncate_to_ms=True)
        raw_index_timeout_ms = int(datetime_to_ms_timestamp(now) - (raw_index_timeout_minutes * 60 * 1000))
        query = {"type": "raw", "timestamp": {"$lt": raw_index_timeout_ms}}
        ret = self.ack_client.delete_many(query)
        _logger.info(f"Deleted {ret.deleted_count} raw indexes in orgsqs table entries")
